import click
import nmtutil
import os
import re
import sys
from common.system import validate_ip_address , run_shell_command
from shutil import copy
from collections import OrderedDict

import logging

logger = logging.getLogger(__name__)

unknown='UNKNOWN'

@click.group(help="Manage network configuration.")
@nmtutil.pass_context
def nmtmain(ctx, **kwargs):
    pass


@nmtmain.command(help="Set network configuration")
@click.option("--device", type=click.STRING, required=False, default=None, help="Network device name")
@click.option("--type", type=click.Choice(["static", "dhcp"]), required=False, default=None, help="Boot protocol")
@click.option("--ip", type=click.STRING, required=False, default=None, help="IP address")
@click.option("--prefix", type=click.INT, required=False, default=None, help="Network prefix [length]")
@click.option("--dns", type=click.STRING, required=False, default=None, help="DNS server")
@click.option("--gateway", type=click.STRING, required=False, default=None, help="Gateway server")
@click.option("--domain", type=click.STRING, required=False, default=None, help="Domain search list")
@click.option("--restart", is_flag=True, help="Restart Network service")

@nmtutil.pass_context
def set(ctx, device, type, ip, prefix, dns, gateway, domain, restart):

    logger.info ("Setting network configuration, device=%s, type=%s, ip=%s, prefix=%s, dns=%s, gateway=%s, domain=%s, restart=%s" 
                   % (device, type, ip, prefix, dns, gateway, domain, restart))

    if not device:
        device,device_name=guess_iface()
    else:
        # TODO
        device_name=device
        
    dict = load_config( device_name )


    bootproto='none'
    ip_validation      = True
    prefix_validation  = True
    dns_validation     = True
    gateway_validation = True

    if type is None or type == 'static': 
        bootproto='none'
    else:
        bootproto='dhcp'

    dict['BOOTPROTO']=bootproto
    
    if bootproto == 'dhcp':
        save_config(dict, device_name )
        # if bootpro is DHCP , all network settings  ignored
        logger.info ( 'The bootproto set to DHCP, reset of network configuration will be ignored') 

        sys.exit(0)


    if domain:
        # DOMAIN string is not validated
        logger.info ( 'Set domain: ' + str(domain)  ) 
        dict['DOMAIN']=domain

    if ip:
        if ctx.verbose: print ( '\tValidaing IP address: ' + str(ip))

        ip_validation=validate_ip_address( str(ip))

        if ip_validation:
            dict['IPADDR']=str(ip)
        else:
            print ( 'IP address ' + str(ip) + ' is invalid.' ) 
            logger.error ( 'IP address ' + str(ip) + ' is invalid.' ) 

    if gateway:
        if ctx.verbose: print ( '\tValidating Gateway IP address: ' + str(gateway))

        gateway_validation=validate_ip_address( str(gateway))

        if gateway_validation:
            dict['GATEWAY']=str(gateway)
        else:
            print('Gateway IP is invalid : ' + str(gateway))
            logger.error('Gateway IP is invalid : ' + str(gateway))


    if prefix:
        if ctx.verbose: print ( '\tValidating Prefix Length: ' + str(prefix))                           

        prefix_validation=validate_ip_address( str(ip) + '/' + str(prefix) )

        if prefix_validation:
             dict['PREFIX']=str(prefix)
        else:
            print('IP or/and Network Prefix is invalid : ' + str(ip) + '/' + str(prefix) )
            logger.error('IP or/and Network Prefix is invalid : ' + str(ip) + '/' + str(prefix))


    if dns:
        if ctx.verbose: print ( '\tValidating IP address: ' + str(dns))
        dns_list = [x.strip() for x in dns.split(',')]

        
        for i in range(len(dns_list) ):

            if ctx.verbose: print ( '\tValidating DNS[' + str(i) + '] = ' + dns_list[i])
            dns_validation=dns_validation and validate_ip_address( dns_list[i] )
            if dns_validation:
                dict['DNS' + str(i+1) ] = dns_list[i]

        if not dns_validation:
            print('One or more of DNS IP is invalid : ' + str(dns) )
            logger.error('One or more of DNS IP is invalid : ' + str(dns) )

    if  ip_validation and prefix_validation and dns_validation and gateway_validation:
        print ( 'All validation passed' )

    else:
        print ( 'Validation failed, network interface remains unchanged.' )
        logger.error ( 'Validation failed, network interface remains unchanged.' )
        sys.exit(1)
        
    save_config(dict, device_name ) 

    if restart:
        logger.info ( 'Restarting network service in order to changes get effect immediately.' )
        output = run_shell_command("systemctl restart network", use_sudo=True )[1]



@nmtmain.command(help="Get network configuration")
@click.option("--device", type=click.STRING, required=False, default=None, help="Network device name")
@click.option("--ip", type=click.STRING, required=False, default=None, help="IP address")
@nmtutil.pass_context
def get(ctx, device, ip):
    #click.echo("Getting network configuration")


    # get domain search 
    domain=unknown
    rc, output = run_shell_command("cat /etc/resolv.conf" , ignore_error=False )
    if not rc:
        for line in output:
            if line.startswith('search'):
                domain=line[7:]



    if not device:
        device,device_name=guess_iface()

    dict = load_config(device)

        
    if device:
        #output = run_shell_command("nmcli connection show --active")[1]
        rc, output = run_shell_command("nmcli device show %s"  %(device), ignore_error=False )

        print("BOOTPROTOCOL:" + '\t'*4 + dict["BOOTPROTO"])
        print( *output, sep = "\n")

        if rc:
            sys.exit(rc)


    elif ip:
        iface=find_iface_by_ip(ip)
        print (iface)

    else:
        rc, output = run_shell_command("nmcli device show")

        print("BOOTPROTOCOL:" + '\t'*4 + dict["BOOTPROTO"])
        print( *output, sep = "\n")

        if rc:
            sys.exit(rc)



#
# not in use
#
def load_network_object(unformated_list):


    network_dict=OrderedDict()
    for line in unformated_list:
        list=re.split(r':[ ]+',line )

        if len(list) == 2:
            network_dict[ list[0] ]=list[1]
        else:
            network_dict[ list[0] ]=''
 
    print ( network_dict )

    return network_dict
#
# returns iface name , if single iface is connected,
# otherwise returns list: device,device_name . device and device name can be different  
#
def guess_iface():
    
    guess=unknown

    # show all connected devices
    output = run_shell_command("nmcli con show --active")[1]

    
    iface = []

    for line in output:
        if not ' ethernet ' in line : continue
        
        # look by device_name, device name can contain_space
        device=re.split('  ',line)[-1]
        device_name=re.split('  ',line)[0]

        #print (device_name + ' , ' + device)

        iface.append(device)
        iface.append(device_name)
        logger.info('Seeing device: ' + device + ' ,  device name: ' + device_name )


    

    #if len(iface) == 0 or len(iface) > 2:
    if len(iface) == 0 : 
        logger.error ( 'Can not detect connected interface, pass  --device=<> option') 
        sys.exit(1)
    else:
        # use first available
        # TODO: revisit
        logger.warning( 'Multiple interfaces encountered , will use first one ')
                    
    logger.info('Guessed network device: ' + iface[0] + ' ,  device name: ' + iface[1] )

    # return only two elements
    return iface[:2]
        
#
# in case multiple iface cards/IP are set
#
def find_iface_by_ip( ip ):
    
    # show all connected devices
    output = run_shell_command("nmcli con show")[1]
    iface = []
    for line in output:
        if line.startswith('NAME'): continue
        iface.append( line.split()[0] )


    found_device=''
    for dev in iface:
        output2 = run_shell_command("nmcli device show %s"  %(dev) )[1]
        
        for line in output2:
            if line.startswith('IP4.ADDRESS') or line.startswith('IP6.ADDRESS'): 
                found_ip=line.split()[1]

                if found_ip.startswith(ip):
                    
                    #print (found_ip + ' appears in device ' + dev)
                    
                    found_device = dev
                    break
    if not found_device:
            logger.warning( 'Cannot find device from IP address ' + ip)
    else:
        logger.info( 'Found device  ' + found_device + ' from IP address ' + ip)
    return found_device

#
# saving confuration in file from dictionary 
#
def save_config ( dict, device_name ):

    device_name_converted=device_name.replace(' ','_')


    interface_config="/etc/sysconfig/network-scripts/ifcfg-" + device_name_converted
    #interface_config="/tmp/ifcfg-" + device

    if os.path.exists ( interface_config ):
        copy( interface_config , interface_config + '.bak' )

    print('Saving configuration in ' + interface_config)
    logger.info('Saving configuration in ' + interface_config )


    if False:
        print ('\nModified network interface:\n' )
        for key in dict: print ( key + '=' + dict[key])


    with open(interface_config, 'w') as out:
        for key in dict: 
            out.write ( key + '=' + dict[key] + '\n')


#
# loading confuration in dictionary 
#   network script name is derived from device name
#
def load_config(device_name='ens160'):

    device_name_converted=device_name.replace(' ','_')


    interface_config="/etc/sysconfig/network-scripts/ifcfg-" + device_name_converted
    if not os.path.exists ( interface_config ):
        print  ( '\nFile ' + interface_config + ' does not exist.' )
        logger.error ( '\nFile ' + interface_config + ' does not exist.' )        
        sys.exit(1)


    net_interface = run_shell_command("cat %s"  %(interface_config) , use_sudo=True )[1]

    dict = {}
    for line in net_interface:
        line = line.strip()
        # If condition to ignore empty lines.
        if len(line) != 0:
            key, value = line.split("=")
            dict[key] = value

    if False:
        print ('Content of original' + interface_config )
        for key in dict: print ( key + '=' + dict[key])

    return dict
